/**
 * Copyright (c) FOM-Nikhef 2014-
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *
 *  Authors:
 *  2014-
 *     Mischa Sall\'e <msalle@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *     <grid-mw-security@nikhef.nl>
 *
 */

/* Define both _XOPEN_SOURCE for strdup and _GNU_SOURCE, this makes Solaris
 * happier */
#define _XOPEN_SOURCE	500
/* Try to provide RTLD_DEFAULT */
/* MacOS needs no POSIX or _DARWIN_C_SOURCE to have RTLD_DEFAULT */
#ifdef __APPLE__
# define _DARWIN_C_SOURCE
#else
# define _GNU_SOURCE
#endif
#include <dlfcn.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pwd.h>

#include <openssl/x509.h>

#include "lcmaps_plugins_robot_config.h"

#include <lcmaps/lcmaps_modules.h>
#include <lcmaps/lcmaps_arguments.h>
#include <lcmaps/lcmaps_cred_data.h>

#if defined(HAVE_LCMAPS_LCMAPS_PLUGIN_PROTOTYPES_H)
#   include <lcmaps/lcmaps_plugin_prototypes.h>
#else
#   include "lcmaps_plugin_prototypes.h"
#endif

#include "lcmaps_gridmapfile.h"
#include "lcmaps_subuserdn.h"


/************************************************************************
 * defines
 ************************************************************************/

#define PLUGIN_PREFIX	"lcmaps_robot_localaccount"

#define PLUGIN_RUN	0   /* full run mode */
#define PLUGIN_VERIFY	1   /* verify-only mode */


/************************************************************************
 * global variables
 ************************************************************************/

static char *gridmapfile = NULL; /* filename of gridmapfile */


/************************************************************************
 * private prototypes
 ************************************************************************/

/* called by plugin_run() and plugin_verify() */
static int plugin_run_or_verify(int argc, lcmaps_argument_t *argv,
				int lcmaps_mode);


/************************************************************************
 * public functions
 ************************************************************************/

/**
 * Initialize function for plugin
 */
int plugin_initialize(int argc, char **argv) {
    const char * logstr = PLUGIN_PREFIX"-plugin_initialize()";
    int i;

    /* Log commandline parameters on debug */
    lcmaps_log(LOG_DEBUG,"%s: passed arguments:\n",logstr);
    for (i=0; i < argc; i++)
	lcmaps_log(LOG_DEBUG,"%s: arg %d is %s\n", logstr, i, argv[i]);

    /* Parse arguments, argv[0] = name of plugin, so start with i = 1 */
    for (i = 1; i < argc; i++) {
	/* check grid-mapfile option (or similar variations) */
        if ( strcmp(argv[i], "-gridmapfile") == 0 ||
             strcmp(argv[i], "-GRIDMAPFILE") == 0 ||
             strcmp(argv[i], "-gridmap") == 0 ||
             strcmp(argv[i], "-GRIDMAP") == 0 )
	{
	    /* check valid filename argument */
            if (argv[i + 1] == NULL || argv[i + 1][0]=='\0') {
		lcmaps_log(LOG_ERR,
		    "%s: option %s needs to be followed by valid filename\n",
		    logstr, argv[i]);
		goto fail_init_module;
	    }
	    /* free existing one and copy new one */
	    free(gridmapfile); gridmapfile=NULL;
	    if (argv[i+1][0]=='/')  { /* absolute path */
		if ( (gridmapfile = strdup(argv[i + 1])) == NULL)	{
		    lcmaps_log(LOG_ERR, "%s: out of memory\n", logstr);
		    goto fail_init_module;
		}
	    } else { /* relative path */
		if (lcmaps_get_prefixed_file(argv[i + 1], &gridmapfile)==-1)
		    goto fail_init_module;
	    }
	    /* log the name and increase arg counter */
	    lcmaps_log(LOG_DEBUG, "%s: Using grid-mapfile \"%s\".\n",
		    logstr, gridmapfile);
            i++;
        }
	/* any other argument is an error */
	else
	{
            lcmaps_log(LOG_ERR,
		    "%s: Unknown argument for plugin: %s (failure)\n",
		    logstr, argv[i]);
            goto fail_init_module;
        }
    }

    /* Check we have a grid-mapfile */
    if (gridmapfile==NULL || gridmapfile[0]=='\0')  {
	free(gridmapfile); gridmapfile=NULL;
        lcmaps_log(LOG_INFO,
	    "%s: No grid-mapfile was provided, will use default.\n", logstr);
    }

    return LCMAPS_MOD_SUCCESS;

fail_init_module:
    free(gridmapfile);
    gridmapfile = NULL;

    return LCMAPS_MOD_FAIL;
}

/**
 * Introspect function for plugin, defines which lcmaps framework data is made
 * available as commandline options for the plugin_run.
 */
int plugin_introspect(int *argc, lcmaps_argument_t **argv) {
    const char * logstr = PLUGIN_PREFIX"-plugin_introspect()";
    static lcmaps_argument_t argList[] = {
	{"px509_chain"  ,"STACK_OF(X509) *", 0,NULL},
	{"pem_string"   ,"char *"          , 0,NULL},
	{"requested_uid","uid_t"           , 0,NULL},
	{NULL           ,NULL              ,-1,NULL},
	{NULL           ,NULL              ,-1,NULL}
    };

    /* Get the version of LCMAPS being used: we need at least 1.5.8 to be able
     * to demand "requested_username" in the argList */
    int major=0,minor=0,patch=0;
    /* Most UNIX now support RTLD_DEFAULT (POSIX reserved) */
#ifdef RTLD_DEFAULT
    int (*lcmaps_major)(void),(*lcmaps_minor)(void),(*lcmaps_patch)(void);
    dlerror();
    lcmaps_major=dlsym(RTLD_DEFAULT,"lcmaps_get_major_version");
    lcmaps_minor=dlsym(RTLD_DEFAULT,"lcmaps_get_minor_version");
    lcmaps_patch=dlsym(RTLD_DEFAULT,"lcmaps_get_patch_version");
    if (dlerror()==NULL)    {
	major=lcmaps_major();
	minor=lcmaps_minor();
	patch=lcmaps_patch();
    }
#else
    /* No RTLD_DEFAULT, just hope the symbol exists in LCMAPS */
    major=lcmaps_get_major_version();
    minor=lcmaps_get_minor_version();
    patch=lcmaps_get_patch_version();
#endif

    /* Too old when older than 1.5.8 */
    if (major<1 || (major==1 && (minor<5 || (minor==5 && patch<8))))	{
	lcmaps_log(LOG_DEBUG,
	    "%s: Old LCMAPS found (%d.%d.%d), not using requested_username\n",
	    logstr,major,minor,patch);
    } else {
	lcmaps_log(LOG_DEBUG,
	    "%s LCMAPS (%d.%d.%d) supports using requested_username\n",
	    logstr,major,minor,patch);
	argList[3].argName="requested_username";
	argList[3].argType="char *";
	argList[3].argInOut=1;
	argList[3].value=NULL;
    }

    lcmaps_log(LOG_DEBUG,"%s: introspecting\n", logstr);

    *argv = argList;
    *argc = lcmaps_cntArgs(argList);
    lcmaps_log(LOG_DEBUG,"%s: address first argument: %p\n",
	    logstr, (void*)argList);

    return LCMAPS_MOD_SUCCESS;
}

/**
 * run function for plugin, wrapper around plugin_run_or_verify() with
 * PLUGIN_RUN mode
 */
int plugin_run(int argc, lcmaps_argument_t *argv) {
    return plugin_run_or_verify(argc, argv, PLUGIN_RUN);
}

/**
 * run function for plugin, wrapper around plugin_run_or_verify() with
 * PLUGIN_VERIFY mode
 */
int plugin_verify(int argc, lcmaps_argument_t *argv) {
    return plugin_run_or_verify(argc, argv, PLUGIN_VERIFY);
}

/**
 * terminate plugin, frees used memory.
 * return LCMAPS_MOD_SUCCESS
 */
int plugin_terminate(void) {
    const char * logstr = PLUGIN_PREFIX"-plugin_terminate()";

    lcmaps_log(LOG_DEBUG,"%s: terminating\n", logstr);

    free(gridmapfile);
    gridmapfile=NULL;

    return LCMAPS_MOD_SUCCESS;
}


/************************************************************************
 * private functions
 ************************************************************************/

/**
 * Actual run/verify function. Called by both plugin_run and plugin_verify
 * with different lcmaps_mode.
 */
static int plugin_run_or_verify(int argc, lcmaps_argument_t *argv,
				int lcmaps_mode) {
    const char *        logstr       = NULL;
    const char *        logmapfile   = gridmapfile ? gridmapfile
						   : "default grid-mapfile";
    void *              value        = NULL;
    STACK_OF(X509)*     chain        = NULL;
    char *              pem          = NULL;
    char *              dn           = NULL;
    char *              req_username = NULL;
    int                 req_username_needs_free=0;
    uid_t               req_uid      = (uid_t)(-1);
    struct passwd *     pw           = NULL;
    unsigned short      options      = 0;
    char *              searchstr    = NULL;
    int                 rc           = 0;
    char *              username     = NULL;
    struct passwd       *user_info   = NULL;
    int                 cnt_sec_gid  = 0;
    gid_t *             sec_gid      = NULL;
    int                 i            = 0;

    /* Set suitable logstr */
    if (lcmaps_mode == PLUGIN_RUN)
        logstr = PLUGIN_PREFIX"-plugin_run()";
    else if (lcmaps_mode == PLUGIN_VERIFY)
        logstr = PLUGIN_PREFIX"-plugin_verify()";
    else {
        lcmaps_log(LOG_ERR, PLUGIN_PREFIX"-plugin_run_or_verify(): "
		"attempt to run plugin in invalid mode: %d\n", lcmaps_mode);
        goto fail_plugin;
    }

    /* Try to get chain from LCMAPS */
    value=lcmaps_getArgValue("px509_chain", "STACK_OF(X509) *", argc, argv);
    if (value == NULL || (chain = *(STACK_OF(X509) **)value) == NULL ) {
	/* No valid chain found in LCMAPS framework, try to obtain from PEM
	 * string */
	lcmaps_log(LOG_DEBUG, "%s: no X.509 chain is set, trying pem string.\n",
		logstr);
	value=lcmaps_getArgValue("pem_string", "char *", argc, argv);
	if (value==NULL || (pem=*(char**)value) == NULL ) {
	    /* also not found: fatal error */
	    lcmaps_log(LOG_WARNING,
		    "%s: no chain or pemstring is set.\n", logstr);
	    return -1;
	}

	/* Convert pem string to chain */
	if (pem_string_to_x509_chain(&chain, pem)!=0)	{
	    lcmaps_log(LOG_WARNING,
		    "%s: cannot convert pemstring to chain.\n", logstr);
	    return -1;
	}
    }

    /* Get per-user sub-proxy DN from certificate chain */
    if (lcmaps_get_subuser_dn(&dn, chain)==-1)   {
	lcmaps_log(LOG_NOTICE,
		"%s: cannot get per-user sub-proxy DN from chain\n", logstr);
	goto fail_plugin;
    }

    /* In verify mode, requested account is typically in requested_uid, in run
     * mode, it can be provided as requested_username. Let's just see if either
     * is set. */

    /* Get requested_username value */
    value = lcmaps_getArgValue("requested_username", "char *", argc, argv);
    if ( value != NULL && *(char **)value !=NULL ) {
	/* copy and log resulting account name */
	req_username=*(char **)value;
	lcmaps_log(LOG_DEBUG,"%s: the requested user is %s\n",
		logstr, req_username);
    } else { /* No (valid) requested_username, try requested_uid */
	/* Get requested_uid value */
	value = lcmaps_getArgValue("requested_uid", "uid_t", argc, argv);
	if ( value!=NULL && *(int *)value != -1)  { /* undefined value -> -1 */
	    req_uid = *(uid_t *)value;
	    /* Basic sanity check */
	    if (req_uid == 0) {
		lcmaps_log(LOG_ERR,
			"%s: illegal request for uid == 0 (failure)\n", logstr);
		goto fail_plugin;
	    }
	    /* Get passwd info */
	    if ( (pw = getpwuid(req_uid )) == NULL ||
		 pw->pw_name == NULL )
	    {
		lcmaps_log(LOG_WARNING,
			"%s: the requested uid %d is illegal.\n",
			logstr, (int) req_uid);
		goto fail_plugin;
	    }
	    /* copy resulting account name */
	    if ( (req_username=strdup(pw->pw_name))==NULL ) {
		lcmaps_log(LOG_ERR,"%s: out of memory\n",logstr);
		goto fail_plugin;
	    }
	    /* keep track whether it needs free */
	    req_username_needs_free=1;
	    /* Log the result */
	    lcmaps_log(LOG_DEBUG,"%s: the requested user is %s(%d)\n",
		    logstr, req_username, (int)req_uid);
	} else if (lcmaps_mode == PLUGIN_VERIFY)    {
	    lcmaps_log(LOG_WARNING,
		    "%s: both requested_username and requested_uid are unset, "
		    "need at least one in PLUGIN_VERIFY mode\n", logstr);
	    goto fail_plugin;
	}
    }

    /* Set the matching options */
    if (req_username)	{
	options = MATCH_EXACT;
	searchstr=req_username;
    } else {
	options = 0;
	searchstr=".";
    }

    /* Do lookup for per-user sub-proxy DN in grid-mapfile */
    rc = lcmaps_gridmapfile(gridmapfile, dn, searchstr, options, &username);

    /* parse return value of lcmaps_gridmapfile */
    if (rc==-1)
	/* error */
	goto fail_plugin;
    if (rc==0 && username==NULL)    { /* default also returns rc==0 */
	/* no match, this should be at most a NOTICE */
	if (req_username)
	    lcmaps_log(LOG_NOTICE,
		"%s: No match for requested localaccount %s for user \"%s\", "
		"in %s\n", logstr, req_username, dn, logmapfile);
	else
	    lcmaps_log(LOG_NOTICE,
		"%s: No match for a localaccount for user \"%s\", "
		"in %s\n", logstr, dn, logmapfile);
	goto fail_plugin;
    }

    /* found match: log */
    lcmaps_log(LOG_DEBUG,"%s: found %susername: %s\n",
	    logstr, req_username ? "requested " : "", username);

    /* Convert username to uid, pgid and sgids */

    /* Get account info for found username */
    if ( (user_info = getpwnam(username)) == NULL ) {
	lcmaps_log(LOG_WARNING,
		"%s: no user account found with the name \"%s\"\n",
		logstr, username);
	goto fail_plugin;
    }

    /* Log resulting account */
    lcmaps_log(LOG_DEBUG,"%s: username : %s\n", logstr, user_info->pw_name);
    lcmaps_log(LOG_DEBUG,"%s: user_id  : %d\n", logstr, user_info->pw_uid);
    lcmaps_log(LOG_DEBUG,"%s: group_id : %d\n", logstr, user_info->pw_gid);
    lcmaps_log(LOG_DEBUG,"%s: home dir : %s\n", logstr, user_info->pw_dir);

    /* Add this credential data to the credential data repository in the plugin
     * manager */
    addCredentialData(UID, &(user_info->pw_uid));
    addCredentialData(PRI_GID, &(user_info->pw_gid));

    /* Retrieve secondary group id's */
    if (lcmaps_get_gidlist(username, &cnt_sec_gid, &sec_gid)==0) {
	for (i = 0; i < cnt_sec_gid; i++)
	    addCredentialData(SEC_GID, &(sec_gid[i]));
	free(sec_gid);
    }

    /* success */
    /* Cleanup chain but only if we allocated it */
    if (pem && chain)
	sk_X509_pop_free(chain, X509_free);
    free(dn);
    if (req_username_needs_free)
	free(req_username);
    free(username);

    lcmaps_log(LOG_INFO,"%s: robot_localaccount plugin succeeded\n", logstr);

    return LCMAPS_MOD_SUCCESS;

fail_plugin:
    /* Cleanup chain but only if we allocated it */
    if (pem && chain)
	sk_X509_pop_free(chain, X509_free);
    free(dn);
    if (req_username_needs_free)
	free(req_username);
    free(username);

    lcmaps_log(LOG_INFO,"%s: robot_localaccount plugin failed\n", logstr);

    return LCMAPS_MOD_FAIL;
}
